import tensorflow as tf
import keras_tuner as kt
import numpy as np
import ipdb
import matplotlib.pyplot as plt

from os import makedirs
from os.path import join
from sys import platform
from functools import partial

from modules.datapipeline import TFDataPipeline
from config import MAX_RESP_RATE, MIN_RESP_RATE, FS_RESAMPLE, WINDOW_SIZE,\
        WINDOW_SHIFT

def set_tf_environ(gpu_to_use:list=[0]):
    gpus = tf.config.list_physical_devices('GPU')
    ngpus = len(gpus)
    assert max(gpu_to_use)<ngpus, 'invalid gpu setting'

    for n in gpu_to_use:
        tf.config.set_visible_devices(gpus[n], 'GPU')
    try:
        if ngpus == 1:
            tf.config.experimental.set_memory_growth(gpus[0], True)
        elif ngpus == 2:
            tf.config.experimental.set_memory_growth(gpus[0], True)
            tf.config.experimental.set_memory_growth(gpus[1], True)
    except:
        Warning("Invalid device or cannot modify virtual devices once"\
                "initialized")

def get_int_units(hp, max_n_layers, name, **kwargs):
    units = []
    for n in range(max_n_layers):
        values_list = list(range(n+1, max_n_layers+1))
        units.append(
            hp.Int(name+str(n),
                   parent_name="n_layers",
                   parent_values=values_list,
                   **kwargs
                  )
        )
    return units

def get_choice_units(hp, max_n_layers, name, values, **kwargs):
    units = []
    for n in range(max_n_layers):
        values_list = list(range(n+1, max_n_layers+1))
        units.append(
            hp.Choice(name+str(n),
                      values,
                      parent_name="n_layers",
                      parent_values=values_list,
                      **kwargs
                     )
        )
    return units

def get_float_units(hp, max_n_layers, name, **kwargs):
    units = []
    for n in range(max_n_layers):
        values_list = list(range(n+1, max_n_layers+1))
        units.append(
            hp.Float(name+str(n),
                   parent_name="n_layers",
                   parent_values=values_list,
                   **kwargs
                  )
        )
    return units

def scheduler(epoch, lr, static_epoch=10, rate=0.1):
    if epoch < static_epoch:
        return lr
    else:
        return lr * tf.math.exp(-rate)

class FNN_HyperModel(kt.HyperModel):
    def __init__(self):
        self.model_type  = 'fnn'
        self.n_labels = 1
        self.n_features = 8
        self.window_size = 10

        self.batch_size = 32
        self.dropout = [0.4, 0.0, 0.15]
        self.learning_rate = 1e-4
        self.n_layers = 3
        self.units = [128, 128, 128]
        self.loss_fn = tf.keras.losses.MeanAbsoluteError()

        # Maybe run window_size with keras_tuner over fixed length and re-run
        # with other sequence lengths
        self.hyperparameters = {
            "batch_size":    [8, 32], #Int, step = 32
            "learning_rate": [1e-6, 1e-2], # float, sampling='log'
            "dropout":       [0.0, 0.5], # float, step=0.1
            "units":         [32, 64, 128, 256, 512, 1024, 2048, 4096],
            "n_layers":      [1, 2, 3, 4, 5], # Choice
        }

        self.is_training = True
        self.verbose = False

        set_tf_environ()

    def build(self, hp):
        n_labels = self.n_labels

        self.input_shape = (None, self.n_features)

        # self.batch_size = hp.Int("batch_size",
        #                          min_value=self.hyperparameters['batch_size'][0],
        #                          max_value=self.hyperparameters['batch_size'][1],
        #                          step=8,
        #                         )
        units_hp = self.hyperparameters['units']
        dropout_hp = self.hyperparameters['dropout']
        n_layers_hp = self.hyperparameters['n_layers']
        n_layers = hp.Choice("n_layers", n_layers_hp)
        units = get_choice_units(hp, n_layers_hp[-1],
                                 'dense_unit',
                                 units_hp)
        dropouts = get_float_units(hp, n_layers_hp[-1],
                                       'dropout',
                                       min_value=dropout_hp[0],
                                       max_value=dropout_hp[1],
                                       step=0.1,
                                      )

        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Input(shape=self.input_shape,
                                        batch_size=self.batch_size))
        for n in range(n_layers):
            dense_unit = tf.keras.layers.Dense(units[n], activation='relu')
            model.add(dense_unit)
            model.add(tf.keras.layers.Dropout(dropouts[n]))

        if n_labels>1:
            activation = 'softmax'
        else:
            activation = 'relu'
        model.add(tf.keras.layers.Dense(self.n_labels, activation=activation))

        model.compile(
            optimizer=tf.keras.optimizers.Adam(self.learning_rate),
            loss=self.loss_fn,
            # loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
            # metrics=[tf.keras.metrics.MeanAbsoluteError()]
        )

        self.model = model
        return model

    def fit(self, hp, model, x, y, validation_data=None,  **kwargs):
        def make_ds(x, y):
            ds_x = tf.data.Dataset.from_tensor_slices(x)\
                    .batch(self.batch_size, drop_remainder=True)
            ds_y = tf.data.Dataset.from_tensor_slices(y)\
                    .batch(self.batch_size, drop_remainder=True)
            ds = tf.data.Dataset.zip((ds_x, ds_y))\
                    .shuffle(3000, reshuffle_each_iteration=True)
            return ds

        train_ds = make_ds(x, y)

        if validation_data is not None:
            val_ds = make_ds(*validation_data)
            history = model.fit(train_ds,
                                validation_data=val_ds,
                                verbose=self.verbose,
                                **kwargs
                               )
        else:
            val_ds = None
            history = model.fit(x, y,
                                verbose=self.verbose,
                                batch_size=self.batch_size,
                                **kwargs
                               )
        return history

class LSTM_HyperModel(FNN_HyperModel):
    def __init__(self):
        super().__init__()
        self.units = None
        self.model_type  = 'lstm'
        self.rnn_units = [128, 128, 128]

        # Maybe run window_size with keras_tuner over fixed length and re-run
        # with other sequence lengths
        self.hyperparameters = {
            "batch_size":    [8, 32], #Int, step = 32
            "learning_rate": [1e-6, 1e-2], # float, sampling='log'
            "dropout":       [0.0, 0.5], # float, step=0.1
            # "rnn_units":     [32, 64, 128, 256, 512, 1024, 2048], # Int, 8
            "rnn_units":     [32, 64, 128, 256], # Int, 8
            "n_layers":      [1, 2, 3, 4, 5], # Choice
            "is_bilstm":     [True, False], # Choice
        }

        self.return_sequences = True
        self.is_training = True
        self.verbose = False

        gpus = tf.config.list_physical_devices('GPU')
        tf.config.experimental.set_memory_growth(gpus[0], True)

    def LSTM(self, rnn_units):
        '''
            The requirements to use the cuDNN implementation are:
            activation == tanh
            recurrent_activation == sigmoid
            recurrent_dropout == 0
            unroll is False
            use_bias is True
            Inputs, if use masking, are strictly right-padded.
            Eager execution is enabled in the outermost context.
            internal dropout are recurrent output dropouts and input dropouts are
            input connection dropouts
        '''
        # Stateful runs the end of the current batch onto the next batch
        return tf.keras.layers.LSTM(
            rnn_units,
            return_sequences=self.return_sequences,
            recurrent_initializer='glorot_uniform',
            stateful=False,
        )

    def build(self, hp):
        n_labels = self.n_labels

        self.input_shape = (self.window_size, self.n_features)

        # self.batch_size = hp.Int("batch_size",
        #                          min_value=self.hyperparameters['batch_size'][0],
        #                          max_value=self.hyperparameters['batch_size'][1],
        #                          step=8,
        #                         )
        rnn_hp = self.hyperparameters['rnn_units']
        dropout_hp = self.hyperparameters['dropout']
        n_layers_hp = self.hyperparameters['n_layers']

        is_bilstm = hp.Boolean("is_bilstm", default=True)
        n_layers = hp.Choice("n_layers", n_layers_hp)

        rnn_units = get_choice_units(hp, n_layers_hp[-1],
                                     'rnn_unit',
                                     rnn_hp)
        # dropouts = get_float_units(hp, n_layers_hp[-1],
        dropouts = get_float_units(hp, n_layers,
                                       'dropout',
                                       min_value=dropout_hp[0],
                                       max_value=dropout_hp[1],
                                       step=0.1,
                                      )

        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Input(shape=self.input_shape,
                                        batch_size=self.batch_size))

        for n in range(n_layers):
            unit = self.LSTM(rnn_units[n])
            if is_bilstm:
                unit = tf.keras.layers.Bidirectional(unit)
            model.add(unit)
            model.add(tf.keras.layers.Dropout(dropouts[n]))

        model.add(tf.keras.layers.GlobalMaxPooling1D())
        model.add(tf.keras.layers.Dense(n_labels))

        # self.learning_rate = hp.Float("learning_rate",
        #                               min_value=self.hyperparameters['learning_rate'][0],
        #                               max_value=self.hyperparameters['learning_rate'][1],
                                     # )
        model.compile(
            optimizer=tf.keras.optimizers.Adam(self.learning_rate),
            loss=self.loss_fn,
        )

        self.model = model
        return model

class CNN1D_HyperModel(FNN_HyperModel):
    def __init__(self):
        super().__init__()
        self.units = None
        self.model_type  = 'cnn1d'

        self.n_layers       = 1
        self.filter_units   = [32]
        self.dense_units    = [48]
        self.kernel_sizes   = [2]
        self.max_pool_sizes = [10]
        self.stride_sizes   = [2]

        self.loss_fn = tf.keras.losses.SparseCategoricalCrossentropy()
        # self.metrics = [tf.keras.metrics.CategoricalAccuracy()]

        # Maybe run window_size with keras_tuner over fixed length and re-run
        # with other sequence lengths
        self.hyperparameters = {
            "batch_size":     [8, 32], #Int, step = 32
            "learning_rate":  [1e-6, 1e-2], # float, sampling='log'
            "dropout":        [0.0, 0.5], # float, step=0.1
            "filter_units":   [32, 64, 128, 256], # Int, 8
            "n_layers":       [1, 2, 3, 4, 5], # Choice
            "kernel_sizes":   [1, 2, 3, 4, 5], # Choice
            "max_pool_sizes": [1, 2, 3, 4, 5], # Choice
            "stride_sizes":   [1, 2, 3, 4, 5], # Choice
            "n_layers":       [1, 2, 3, 4, 5], # Choice
        }

        self.is_training = True
        self.verbose = False

        gpus = tf.config.list_physical_devices('GPU')
        tf.config.experimental.set_memory_growth(gpus[0], True)

    def build(self, hp):
        n_labels = self.n_labels

        # self.batch_size = hp.Int("batch_size",
        #                          min_value=self.hyperparameters['batch_size'][0],
        #                          max_value=self.hyperparameters['batch_size'][1],
        #                          step=8,
        #                         )
        filter_hp = self.hyperparameters['filter_units']
        dropout_hp = self.hyperparameters['dropout']
        n_layers_hp = self.hyperparameters['n_layers']
        kernel_hp = self.hyperparameters['kernel_sizes']
        pool_hp = self.hyperparameters['max_pool_sizes']
        stride_hp = self.hyperparameters['stride_sizes']

        n_layers = hp.Choice("n_layers", n_layers_hp)

        filter_units = get_choice_units(hp, n_layers_hp[-1],
                                        'filter_unit',
                                        filter_hp)
        kernel_sizes = get_choice_units(hp, n_layers_hp[-1],
                                        'kernel_size',
                                        kernel_hp)
        pool_sizes = get_choice_units(hp, n_layers_hp[-1],
                                      'pool_size',
                                      pool_hp)
        stride_sizes = get_choice_units(hp, n_layers_hp[-1],
                                        'stride_size',
                                        stride_hp)
        dropouts = get_float_units(hp, n_layers_hp[-1],
                                       'dropout',
                                       min_value=dropout_hp[0],
                                       max_value=dropout_hp[1],
                                       step=0.1,
                                      )

        model = tf.keras.Sequential(name='cnn1d')
        # model.add(
        #     tf.keras.layers.InputLayer(
        #         input_shape=self.input_shape, batch_size=None,
        #         name='inputlayer',
        #     )
        # )
        for n in range(n_layers):
            filt_unit   = filter_units[n]
            kernel_size = kernel_sizes[n]
            pool_size   = pool_sizes[n]
            stride_size = stride_sizes[n]
            dropout = dropouts[n]

            model.add(
                # convolutional layer
                tf.keras.layers.Conv1D(filters=filt_unit,
                                       kernel_size=kernel_size,
                                       activation='relu',
                                       padding='same',
                                       name=f'conv1d_{n}'
                                      )
            )
            model.add( tf.keras.layers.BatchNormalization(name=f'batchnorm_{n}') )
            model.add(
                # max pooling layer, striding across with some filter
                tf.keras.layers.MaxPool1D(pool_size=pool_size,
                                          strides=stride_size,
                                          name=f'maxpool_{n}'
                                         )
            )
            # Dropout layer to avoid overfitting
            model.add( tf.keras.layers.Dropout(dropout, name=f'dropout_{n}') )

        model.add( tf.keras.layers.GlobalMaxPooling1D(name=f'gmp_{n}') )
        if n_labels>1:
            activation = 'softmax'
        else:
            activation = None
        model.add(tf.keras.layers.Dense(n_labels, activation=activation,
                                        name=f'dense_output'))

        model.compile(
            optimizer=tf.keras.optimizers.Adam(self.learning_rate),
            loss=self.loss_fn,
            # metrics=self.metrics,
        )

        self.model = model
        return model

class TunerClass():
    def __init__(self, hypermodel, marker=1, tuner_type='hyperband',
                 overwrite=True, directory=None):
        if platform =='linux' or platform == 'linux2':
            self.sep = "/"
        else:
            self.sep = "\\"

        self.tuner_type = tuner_type

        self.model_type = hypermodel.model_type
        if directory is None:
            self.directory = f'/data/rqchia/ARIA/'\
                   f'{self.model_type}_hypermodel_{marker}'
        else:
            self.directory = directory
        self.project_name = self.tuner_type

        self.monitor = 'val_loss'
        self.overwrite = overwrite

        self.max_epochs = 10
        self.max_trials = 20

        self.tuner = self.set_tuner(hypermodel)
        self.callbacks = self.get_callbacks()

        self.best_model_path = join(self.directory, self.project_name,
                                    'best_model','ckpt')

    def set_tuner(self, hypermodel):
        ''' Tuners here '''
        directory    = self.directory
        project_name = self.project_name
        monitor      = self.monitor
        overwrite    = self.overwrite
        max_epochs   = self.max_epochs
        max_trials   = self.max_trials
        tuner_type   = self.tuner_type

        if tuner_type == 'hyperband':
            tuner = kt.Hyperband(
                objective=monitor,
                max_epochs=max_epochs,
                hyperband_iterations=1,
                hypermodel=hypermodel,
                directory=directory,
                project_name=project_name,
                overwrite=overwrite,
            )
        elif tuner_type == 'randomsearch':
            tuner = kt.RandomSearch(
                objective=monitor,
                max_trials=max_trials,
                hypermodel=hypermodel,
                directory=directory,
                project_name=project_name,
                overwrite=overwrite,
            )
        elif tuner_type == 'bayesianoptimization':
            tuner = kt.BayesianOptimization(
                hypermodel=hypermodel,
                objective=monitor,
                max_trials=max_trials,
                directory=directory,
                project_name=project_name,
                overwrite=overwrite,
            )

        self.tuner = tuner
        return tuner

    def get_callbacks(self, epochs=None):
        ''' Callbacks here '''
        if epochs is None:
            epochs = self.max_epochs

        stop_early = tf.keras.callbacks.EarlyStopping(
            monitor=self.monitor, patience=epochs//10, verbose=1, mode='auto',
            restore_best_weights=True,
        )
        nan_callback = tf.keras.callbacks.TerminateOnNaN()
        schedule_func = partial(scheduler, static_epoch=epochs//2)
        lr_scheduler = tf.keras.callbacks.LearningRateScheduler(
            schedule_func)
        return [stop_early, nan_callback, lr_scheduler]

    def search(self, x, y, validation_data, **kwargs):
        callbacks = self.callbacks[1:]
        self.tuner.search(x, y, validation_data,
                          epochs=self.max_epochs,
                          callbacks=callbacks,
                          use_multiprocessing=True, **kwargs)

    def load_model(self, ind=0, is_training=False):
        tuner = self.tuner
        best_hps = tuner.get_best_hyperparameters(self.max_trials)[ind]
        print(best_hps.values)

        if not is_training:
            tuner.hypermodel.is_training = False
            tuner.hypermodel.batch_size = 1
            best_hps.values['batch_size'] = 1

        model = tuner.hypermodel.build(best_hps)
        self.tuner.hypermodel.model = model

        return model

    def get_trial_path(self, ind=0):
        best_ind = 0
        tuner = self.tuner
        best_trial = tuner.oracle.get_best_trials(
            self.max_trials)[best_ind].trial_id
        best_trial_path = join(self.directory, self.project_name,
                               'trial_' + best_trial)
        return best_trial_path

    def load_select_weights(self, ind=0):
        trial_path = self.get_trial_path(ind=ind)
        tmp = tf.train.latest_checkpoint(trial_path)
        model = self.tuner.hypermodel.model

        print('loading weights...',end=' ')
        model.load_weights(tmp).expect_partial()
        print('done')

        self.tuner.hypermodel.model = model
        return model

    def save_weights_to_path(self, path=None):
        if path is None:
            path = self.best_model_path
        self.tuner.hypermodel.model.save_weights(path)

    def load_weights_from_path(self, path=None):
        if path is None:
            path = self.best_model_path
        self.tuner.hypermodel.model.load_weights(path)
        return self.tuner.hypermodel.model

