# resnet model 
# when tuning start with learning rate->mini_batch_size -> 
# momentum-> #hidden_units -> # learning_rate_decay -> #layers 
import tensorflow.keras as keras
import tensorflow as tf
import numpy as np
import time
import ipdb

import matplotlib.pyplot as plt

from modules.datapipeline import TFDataPipeline

class Classifier_RESNET_2D:
    def __init__(self, output_directory, input_shape, nb_classes,
                 verbose=False, build=True, load_weights=False):
        self.output_directory = output_directory
        self.verbose = verbose
        self.input_shape = input_shape
        self.nb_classes = nb_classes
        self.batch_size = 32
        self.nb_epochs = 50
        self.n_feature_maps = 32
        self.learning_rate = 1e-3
        if build == True:
            self.model = self.build_model(input_shape)
        if (verbose == True):
            self.model.summary()
        if load_weights == True:
            self.model.load_weights(self.output_directory)
        else:
            self.model.save_weights(self.output_directory + 'ckpt')

    def build_model(self, input_shape=None, downsample=2):
        if input_shape is not None: self.input_shape = input_shape
        n_feature_maps = self.n_feature_maps

        img_shape = [int(val//downsample) 
                     for val in self.input_shape[:-1]]

        input_layer = keras.layers.Input(self.input_shape)
        resize_layer = keras.layers.Resizing(*img_shape)(input_layer)

        # BLOCK 1

        conv_x = keras.layers.Conv2D(filters=n_feature_maps, kernel_size=8,
                                     padding='same')(resize_layer)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv2D(filters=n_feature_maps, kernel_size=5,
                                     padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv2D(filters=n_feature_maps, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # expand channels for the sum
        shortcut_y = keras.layers.Conv2D(filters=n_feature_maps,
                                         kernel_size=1, 
                                         padding='same')(resize_layer)
        shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

        output_block_1 = keras.layers.add([shortcut_y, conv_z])
        output_block_1 = keras.layers.Activation('relu')(output_block_1)

        # BLOCK 2

        conv_x = keras.layers.Conv2D(filters=n_feature_maps * 2,
                                     kernel_size=8, padding='same')(output_block_1)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv2D(filters=n_feature_maps * 2,
                                     kernel_size=5, padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv2D(filters=n_feature_maps * 2, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # expand channels for the sum
        shortcut_y = keras.layers.Conv2D(filters=n_feature_maps * 2,
                                         kernel_size=1,
                                         padding='same')(output_block_1)
        shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

        output_block_2 = keras.layers.add([shortcut_y, conv_z])
        output_block_2 = keras.layers.Activation('relu')(output_block_2)

        # BLOCK 3

        conv_x = keras.layers.Conv2D(filters=n_feature_maps * 2, kernel_size=8,
                                     padding='same')(output_block_2)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv2D(filters=n_feature_maps * 2, kernel_size=5,
                                     padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv2D(filters=n_feature_maps * 2, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # no need to expand channels because they are equal
        shortcut_y = keras.layers.BatchNormalization()(output_block_2)

        output_block_3 = keras.layers.add([shortcut_y, conv_z])
        output_block_3 = keras.layers.Activation('relu')(output_block_3)

        # FINAL

        gap_layer = keras.layers.GlobalAveragePooling2D()(output_block_3)

        output_layer = keras.layers.Dense(self.nb_classes,)(gap_layer)

        model = keras.models.Model(inputs=input_layer, outputs=output_layer)

        # model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(),
        #               metrics=['accuracy'])
        model.compile(loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
                      optimizer=keras.optimizers.Adam(self.learning_rate),
                      metrics=['accuracy'])

        reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50, min_lr=0.0001)

        file_path = self.output_directory + 'best_model.hdf5'

        model_checkpoint = keras.callbacks.ModelCheckpoint(filepath=file_path, monitor='loss',
                                                           save_best_only=True)

        self.callbacks = [reduce_lr, model_checkpoint]

        return model

    def fit(self, x_train, y_train, x_val, y_val, reduce_mean=True):
        if not tf.test.is_gpu_available:
            print('error')
            exit()

        batch_size = self.batch_size
        nb_epochs = self.nb_epochs
        # nb_epochs = 150

        mini_batch_size = int(min(x_train.shape[0] / 10, batch_size))
        dpl = TFDataPipeline(window_size=self.input_shape[0],
                             batch_size=mini_batch_size)
        dpl.window_shift = 120
        train_ds_x = dpl.get_dataset(x_train)
        train_ds_y = dpl.get_dataset(y_train, reduce_mean=reduce_mean)
        train_ds = dpl.zip_datasets(train_ds_x, train_ds_y)

        val_ds_x = dpl.get_dataset(x_val)
        val_ds_y = dpl.get_dataset(y_val, reduce_mean=reduce_mean)
        val_ds = dpl.zip_datasets(val_ds_x, val_ds_y)

        start_time = time.time()

        hist = self.model.fit(train_ds, batch_size=mini_batch_size, epochs=nb_epochs,
                              verbose=self.verbose,
                              validation_data=val_ds,
                             )

        duration = time.time() - start_time

        self.model.save(self.output_directory + 'last_model.hdf5')

        # save predictions
        # np.save(self.output_directory + 'y_pred.npy', y_pred)

        # convert the predicted from binary to integer
        # y_pred = np.argmax(y_pred, axis=1)

        keras.backend.clear_session()

        return hist

    def predict(self, x_test, y_test, return_df_metrics=True):
        start_time = time.time()
        model_path = self.output_directory + 'ckpt'
        model = keras.models.load_model(model_path)
        y_pred = model.predict(x_test)
        if return_df_metrics:
            # y_pred = np.argmax(y_pred, axis=1)
            df_metrics = model.evaluate(y_test, y_pred)
            return df_metrics

class Classifier_RESNET:

    def __init__(self, output_directory, input_shape, nb_classes,
                 verbose=False, build=True, load_weights=False):
        self.output_directory = output_directory
        self.verbose = verbose
        self.input_shape = input_shape
        self.nb_classes = nb_classes
        self.batch_size = 32
        self.nb_epochs = 50
        self.n_feature_maps = 32
        if build == True:
            self.model = self.build_model(input_shape)
        if (verbose == True):
            self.model.summary()
        if load_weights == True:
            self.model.load_weights(self.output_directory)
        else:
            self.model.save_weights(self.output_directory + 'ckpt')

    def build_model(self, input_shape):
        n_feature_maps = self.n_feature_maps

        input_layer = keras.layers.Input(input_shape)

        # BLOCK 1

        conv_x = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=8,
                                     padding='same')(input_layer)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=5,
                                     padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # expand channels for the sum
        shortcut_y = keras.layers.Conv1D(filters=n_feature_maps,
                                         kernel_size=1, 
                                         padding='same')(input_layer)
        shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

        output_block_1 = keras.layers.add([shortcut_y, conv_z])
        output_block_1 = keras.layers.Activation('relu')(output_block_1)

        # BLOCK 2

        conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2,
                                     kernel_size=8, padding='same')(output_block_1)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2,
                                     kernel_size=5, padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # expand channels for the sum
        shortcut_y = keras.layers.Conv1D(filters=n_feature_maps * 2,
                                         kernel_size=1,
                                         padding='same')(output_block_1)
        shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

        output_block_2 = keras.layers.add([shortcut_y, conv_z])
        output_block_2 = keras.layers.Activation('relu')(output_block_2)

        # BLOCK 3

        conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=8,
                                     padding='same')(output_block_2)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=5,
                                     padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # no need to expand channels because they are equal
        shortcut_y = keras.layers.BatchNormalization()(output_block_2)

        output_block_3 = keras.layers.add([shortcut_y, conv_z])
        output_block_3 = keras.layers.Activation('relu')(output_block_3)

        # FINAL

        gap_layer = keras.layers.GlobalAveragePooling1D()(output_block_3)

        output_layer = keras.layers.Dense(self.nb_classes,)(gap_layer)
                                          # activation='softmax')(gap_layer)

        model = keras.models.Model(inputs=input_layer, outputs=output_layer)

        # model.compile(loss='categorical_crossentropy', optimizer=keras.optimizers.Adam(),
        #               metrics=['accuracy'])
        model.compile(loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
                      optimizer=keras.optimizers.Adam(),
                      metrics=['accuracy'])

        reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50, min_lr=0.0001)

        file_path = self.output_directory + 'best_model.hdf5'

        model_checkpoint = keras.callbacks.ModelCheckpoint(filepath=file_path, monitor='loss',
                                                           save_best_only=True)

        self.callbacks = [reduce_lr, model_checkpoint]

        return model

    def fit(self, x_train, y_train, x_val, y_val, reduce_mean=True):
        if not tf.test.is_gpu_available:
            print('error')
            exit()

        batch_size = self.batch_size
        nb_epochs = self.nb_epochs
        # nb_epochs = 150

        mini_batch_size = int(min(x_train.shape[0] / 10, batch_size))
        dpl = TFDataPipeline(window_size=self.input_shape[0],
                             batch_size=mini_batch_size)
        dpl.window_shift = 120
        train_ds_x = dpl.get_dataset(x_train)
        train_ds_y = dpl.get_dataset(y_train, reduce_mean=reduce_mean)
        train_ds = dpl.zip_datasets(train_ds_x, train_ds_y)

        val_ds_x = dpl.get_dataset(x_val)
        val_ds_y = dpl.get_dataset(y_val, reduce_mean=reduce_mean)
        val_ds = dpl.zip_datasets(val_ds_x, val_ds_y)

        start_time = time.time()

        hist = self.model.fit(train_ds, batch_size=mini_batch_size, epochs=nb_epochs,
                              verbose=self.verbose,
                              validation_data=val_ds,
                             )

        duration = time.time() - start_time

        self.model.save(self.output_directory + 'last_model.hdf5')

        # save predictions
        # np.save(self.output_directory + 'y_pred.npy', y_pred)

        # convert the predicted from binary to integer
        # y_pred = np.argmax(y_pred, axis=1)

        keras.backend.clear_session()

        return hist

    def predict(self, x_test, y_test, return_df_metrics=True):
        start_time = time.time()
        model_path = self.output_directory + 'ckpt'
        model = keras.models.load_model(model_path)
        y_pred = model.predict(x_test)
        if return_df_metrics:
            # y_pred = np.argmax(y_pred, axis=1)
            df_metrics = model.evaluate(y_test, y_pred)
            return df_metrics

class Regressor_RESNET:

    def __init__(self, output_directory, input_shape,
                 verbose=False, build=True, load_weights=False):
        self.output_directory = output_directory
        self.verbose = verbose
        self.input_shape = input_shape
        if build == True:
            self.model = self.build_model(input_shape)
        if (verbose == True):
            self.model.summary()
        if load_weights == True:
            self.model.load_weights(self.output_directory)
        else:
            self.model.save_weights(self.output_directory + 'ckpt')

    def build_model(self, input_shape):
        n_feature_maps = 64

        input_layer = keras.layers.Input(input_shape)

        # BLOCK 1

        conv_x = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=8,
                                     padding='same')(input_layer)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=5,
                                     padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv1D(filters=n_feature_maps, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # expand channels for the sum
        shortcut_y = keras.layers.Conv1D(filters=n_feature_maps,
                                         kernel_size=1, 
                                         padding='same')(input_layer)
        shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

        output_block_1 = keras.layers.add([shortcut_y, conv_z])
        output_block_1 = keras.layers.Activation('relu')(output_block_1)

        # BLOCK 2

        conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2,
                                     kernel_size=8, padding='same')(output_block_1)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2,
                                     kernel_size=5, padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # expand channels for the sum
        shortcut_y = keras.layers.Conv1D(filters=n_feature_maps * 2,
                                         kernel_size=1,
                                         padding='same')(output_block_1)
        shortcut_y = keras.layers.BatchNormalization()(shortcut_y)

        output_block_2 = keras.layers.add([shortcut_y, conv_z])
        output_block_2 = keras.layers.Activation('relu')(output_block_2)

        # BLOCK 3

        conv_x = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=8,
                                     padding='same')(output_block_2)
        conv_x = keras.layers.BatchNormalization()(conv_x)
        conv_x = keras.layers.Activation('relu')(conv_x)

        conv_y = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=5,
                                     padding='same')(conv_x)
        conv_y = keras.layers.BatchNormalization()(conv_y)
        conv_y = keras.layers.Activation('relu')(conv_y)

        conv_z = keras.layers.Conv1D(filters=n_feature_maps * 2, kernel_size=3,
                                     padding='same')(conv_y)
        conv_z = keras.layers.BatchNormalization()(conv_z)

        # no need to expand channels because they are equal
        shortcut_y = keras.layers.BatchNormalization()(output_block_2)

        output_block_3 = keras.layers.add([shortcut_y, conv_z])
        output_block_3 = keras.layers.Activation('relu')(output_block_3)

        # FINAL

        gap_layer = keras.layers.GlobalAveragePooling1D()(output_block_3)

        output_layer = keras.layers.Dense(1,
                                          activation='relu')(gap_layer)

        model = keras.models.Model(inputs=input_layer, outputs=output_layer)

        model.compile(loss='mean_squared_error', optimizer=keras.optimizers.Adam(),
                      metrics=['accuracy'])

        reduce_lr = keras.callbacks.ReduceLROnPlateau(monitor='loss', factor=0.5, patience=50, min_lr=0.0001)

        file_path = self.output_directory + 'best_model.hdf5'

        model_checkpoint = keras.callbacks.ModelCheckpoint(filepath=file_path, monitor='loss',
                                                           save_best_only=True)

        self.callbacks = [reduce_lr, model_checkpoint]

        return model

    def fit(self, x_train, y_train, x_val, y_val, reduce_mean=False):
        if not tf.test.is_gpu_available:
            print('error')
            exit()

        batch_size = 32
        # nb_epochs = 150
        nb_epochs = 20

        mini_batch_size = int(min(x_train.shape[0] / 10, batch_size))
        dpl = TFDataPipeline(window_size=self.input_shape[0],
                             batch_size=mini_batch_size)
        dpl.shuffle_flag = False
        dpl.window_shift = 120 # 1 second
        train_ds_x = dpl.get_dataset(x_train)
        train_ds_y = dpl.get_dataset(y_train, reduce_mean=reduce_mean)
        train_ds = dpl.zip_datasets(train_ds_x, train_ds_y)

        val_ds_x = dpl.get_dataset(x_val)
        val_ds_y = dpl.get_dataset(y_val, reduce_mean=reduce_mean)
        val_ds = dpl.zip_datasets(val_ds_x, val_ds_y)

        start_time = time.time()

        print("fitting model")
        hist = self.model.fit(train_ds, batch_size=mini_batch_size, epochs=nb_epochs,
                              verbose=self.verbose,
                              validation_data=val_ds,
                             )

        duration = time.time() - start_time

        self.model.save(self.output_directory + 'last_model.hdf5')

        # save predictions
        # np.save(self.output_directory + 'y_pred.npy', y_pred)

        # convert the predicted from binary to integer
        # y_pred = np.argmax(y_pred, axis=1)

        keras.backend.clear_session()

        return hist

    def predict(self, x_test, y_test, return_df_metrics=True):
        start_time = time.time()
        model_path = self.output_directory + 'ckpt'
        model = keras.models.load_model(model_path)
        y_pred = model.predict(x_test)
        if return_df_metrics:
            # y_pred = np.argmax(y_pred, axis=1)
            df_metrics = model.evaluate(y_test, y_pred)
            return df_metrics
